/*
** Copyright (c) Massachusetts Institute of Technology 1994, 1995, 1996.
**          All Rights Reserved.
**          Unpublished rights reserved under the copyright laws of
**          the United States.
**
** THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
** OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
**
** This code is distributed freely and may be used freely under the 
** following conditions:
**
**     1. This notice may not be removed or altered.
**
**     2. This code may not be re-distributed or modified
**        without permission from MIT (contact 
**        lclint-request@larch.lcs.mit.edu.)  
**
**        Modification and re-distribution are encouraged,
**        but we want to keep track of changes and
**        distribution sites.
*/
/*
** flagMarker.c
*/

# include "lclintMacros.nf"
# include "basic.h"

flagMarker flagMarker_createLocalSet (flagcode code, ynm set, fileloc loc)
{
  flagMarker c = (flagMarker) dmalloc (sizeof (*c));
  
  c->kind = FMK_LOCALSET;
  c->code = code;
  c->info.set = set;
  c->loc = fileloc_copy (loc); 

  return c;
}

flagMarker flagMarker_createSuppress (flagcode code, fileloc loc)
{
  flagMarker c = (flagMarker) dmalloc (sizeof (*c));
  
  c->kind = FMK_SUPPRESS;
  c->code = code;
  c->loc = fileloc_copy (loc); 

  return c;
}

flagMarker flagMarker_createIgnoreOn (fileloc loc)
{
  flagMarker c = (flagMarker) dmalloc (sizeof (*c));
  
  c->kind = FMK_IGNOREON;
  c->code = INVALID_FLAG;
  c->loc = fileloc_copy (loc); 

  return c;
}

flagMarker flagMarker_createIgnoreCount (int count, fileloc loc)
{
  flagMarker c = (flagMarker) dmalloc (sizeof (*c));
  
  c->kind = FMK_IGNORECOUNT;
  c->code = INVALID_FLAG;
  c->info.nerrors = count;
  c->loc = fileloc_copy (loc); 

  return c;
}

flagMarker flagMarker_createIgnoreOff (fileloc loc)
{
  flagMarker c = (flagMarker) dmalloc (sizeof (*c));
  
  c->kind = FMK_IGNOREOFF;
  c->code = INVALID_FLAG;
  c->loc = fileloc_copy (loc); 

  return c;
}

ynm flagMarker_getSet (flagMarker f)
{
  llassert (f->kind == FMK_LOCALSET);

  return f->info.set;
}

flagcode flagMarker_getCode (flagMarker f)
{
  llassert (f->kind == FMK_LOCALSET|| f->kind == FMK_SUPPRESS);

  return f->code;
}

int flagMarker_getCount (flagMarker f)
{
  llassert (f->kind == FMK_IGNORECOUNT);

  return f->info.nerrors;
}

cstring flagMarker_unparse (flagMarker c)
{
  switch (c->kind)
    {
    case FMK_LOCALSET:
      return (message ("%q: %s%s", 
		       fileloc_unparse (c->loc), ynm_unparseCode (c->info.set), 
		       flagcode_name (c->code)));
    case FMK_IGNORECOUNT:
      return (message ("%q: ignore count %d", 
		       fileloc_unparse (c->loc), c->info.nerrors));
    case FMK_IGNOREON:
      return (message ("%q: ignore on", 
		       fileloc_unparse (c->loc)));
    case FMK_IGNOREOFF:
      return (message ("%q: ignore off", 
		       fileloc_unparse (c->loc)));
    case FMK_SUPPRESS:
      return (message ("%q: suppress %s", 
		       fileloc_unparse (c->loc),
		       flagcode_name (c->code)));
    }
}
  
void flagMarker_free (/*@only@*/ flagMarker c)
{
  sfree (c);
}

bool flagMarker_sameFile (flagMarker c, fileloc loc)
{
  return (fileloc_almostSameFile (c->loc, loc));
}

/*
** return true if loc is before c->loc
*/

bool flagMarker_beforeMarker (flagMarker c, fileloc loc)
{
  return  (!fileloc_notAfter (c->loc, loc));
}



